#!/usr/bin/python3

import time
from numpy import array as np_array

import logging

from nx5d.h5like.tools import *

__all__ = [ "ScanReader" ]

'''
The Scan-Streak-Frame concept uses the following HDF5 API elements
to implement sophisticated iteration and slicing helpers:

  - Accessing of data using slicers and the indexing operator (`__getitem__`).

  - `__getitem__` access works both for selecting paths / subpaths, or selecting
    data containers within paths, or slicing parts of the data container itself,
    i.e. sub-access to limited parts of the images (i.e. only specific
    region-of-interests) via slicing. Examples:
    ```
      data["detector"] -> directory containing the "acme" subdirectory

      data["detector/acme"] -> directory, containing the "images" dataset
      data["detector"]["acme"] -> same as above

      data["detector/acme/images"] -> the dataset itself
      data["detector"]["acme"]["images"] -> same as above

      data["detector/acme/images"][3] -> image no. 3 from the dataset

      data["detector/acme/images"][3][60:160,60:160] -> a 100x100 pixel area centered
                                                        around pixel 110x110 of image no. 3

    ```

  - The HDF5 taxonomy (i.e. the '.attrs' dictionary

  - We're not using any HDF5 specific properties or Python attributes.
'''
    
class ScanReader:
    '''
    Loads data from one X-ray diffraction (XRD) scan run and offers
    convesion of images to Q space for further analysis.
    
    The idea is to open/initialize metadata (experimental parameters,
    ROIs, sizes, ...) on construction, and then subsequently "pull out"
    image data. This is supposed to work on a "live" file, i.e. one
    being measured and having data added to it.

    It assumes that data is available in a HDF5-like interface.
    It opens/closes the data object with every use to account for
    parallel use by data acquistion and/or other consumers.

    The assumptions to the data are, essentially (see also `nx5d.xrd.kmc3`):

      - All relevant datasets (frame images, angles etc) are organized
        as (arrays) of datasets with the first coordinate corresponding
        to the progression of frames, i.e. Nx(WxH) for images N for a
        particular angle etc.

      - The frame number progression (index starting with 0) corresponds
        to one or two experimental scan parameters, i.e. time delay (for
        pump-probe experiments), or a particular tilt angle. The `Chunk`
        object reflects this.

      - Instrument definition and angles are compatible with `xrayutilities`,
        the package used for the Q-space conversion.

      - The data has a Nexus-like / ESRF-style layout; this means that
        all relevant measurement data is under a specific folder of type
        "NXcollection" (typically called "measurement", but this is
        configurable), and all relevant parameterisation is stored under
        another folder (typically "instrument/positioners", but again,
        configurable). Data specification is expected relative to these
        folders.

      - Quite generally, the assumption is that we're interested in one or
        more measurement sets, and each of those depends on a subset of
        positioners.
    
    '''
    
    def __init__(self, H5pyLike, scan,
                 h5kwargs=None,
                 pathKeys=None,
                 **dreq):
        '''
        Initializes the object and basic experimental parameters.
        Needs the following:

          - `H5pyLike`: A class or a factory which generates a `h5py`-like object for
            data access. Defaults to `h5py.File`, but objects with a similar API
            (e.g. the `silx.io.spech5.SpecH5` kind of loaders) are also accepted.        

          - `scan`: A parameter list that contains information to be passed to
            `H5pyLike` in order to locate the scan within the data container.
            If the first item in that list is a string (i.e. a file path), then
            the format "resource[#path]" is accepted to pass on both the file name
            ("resource") and the path inside the container ("path"). All other
            items are passed on as unnamed arguments to `H5pyFile`.
            A typical example of this is for instance
            `("/path/to/file.h5#scan23", "r")` if `H5pyLike` is a `h5py.File`
            object. For convenience, if `H5pyLike` is indeed a `h5py.File`.

          - `h5kwargs`: `None` by default, but can be passed a dictionary of named options
            to be passed on to `loader` in addition to the resource locator string.

          - `pathKeys`: Dictionary with shortcuts of standard (?) paths. Currently,
            `ScanReader` recognizes and uses the following paths:
        
              - "measurement": Where the data is

              - "positioners": Where the positioners are (e.g. angles)

          - `dreq`: This is a list of "data request" named parameters, i.e. each key is
            a measureables, to which the value is a list of other measureables or
            poisitioners. The names of the data are relative to `measurablePath`,
            respectively `positionerPath`. The data labels may also contain a path
            component (e.g. "detector/image"), in which case all path separators
            ("/") will be replaced by underscore when data naming is performed
            ("detector_image").

            FIXME: not exactly sure how this is supposed to work. There are two main
            problems with it.
        
             1. The angles: designation of axes (goniometer, detector and image)
                are already in the 'setup' parameter. But the _actual_ angles themselves
                are not. So we have two options here: squeeze them in 'setup' (in which
                case: what is this parameter for?), or squeeze them here (in which case:
                in which order? See also 2...)

             2. The data itself: most of the time the data is the actual detector
                image / frame data. But sometimes we have data of secondary interest,
                e.g. SBcurrent (ring curren) or similar, e.g. for normalization. *This*
                data does not depend on any kind of angles or positioners... or may
                only depend on _some_ positioners. How do we model that?
        '''

        self.pathKeys = pathKeys or {'measurement': 'measurement',
                                     'positioners': 'instrument/positioners',
                                     'instrument': 'instrument' }
        
        self.h5kwargs = h5kwargs or {}

        self.dataSource = self.__source_factory(scan, H5pyLike)

        self.setupTemplate = dreq.get("setup", {})

        self.requiredDatasets = dreq.copy()

        
    def __source_factory(self, params, H5pyLike):
        # Implements magic to parse out the '#path' part of the filename.
        # Returns a callable that creates a usable h5py-like object
        # for querying scan data.

        self.H5pyLike = H5pyLike
        fileUrl, fileArgs = (None, tuple())

        if isinstance(params, str):
            fileUrl = params
        elif isinstance(params, tuple):
            fileUrl, fileArgs = params[0], params[1:]
        else:
            fileUrl = params

        if isinstance(fileUrl, str):
            tmp = fileUrl.split('#')
            if len(tmp) >= 1:
                filePath = tmp[0]
            if len(tmp) == 2:
                self.h5scan = tmp[1]
            else:
                self.h5scan = "/"
            self.h5args = tuple([filePath] + list(fileArgs))
        else:
            self.h5scan = "/"
            self.h5args = (fileUrl,)

        logging.debug("Loading: %r, args: %r, kw: %r" % (self.H5pyLike, self.h5args, self.h5kwargs))

        return lambda: self.H5pyLike(*(self.h5args), **(self.h5kwargs))


    def __defPath(self, locator):
        ''' Translates "syntactic sugar" when specifying certain HDF5 data locators.
        
        Takes a HDF5 locator as accepted by `HdfCarver` and if it's
        a string, or a tuple (str, slice, ...) adds default paths in front
        of the string if necessary (e.g. turns "data" into "@{measurement}/data").

        Everything else is returned unchanged.

        Args:
            locator: eiter a string containing a HDF5 address (`"@..."`), or a tuple
              `("...", slice(), ...)` containing a HDF5 address as its first element
              and only slice objects for every other element.

        Returns:
            The locator, translated to contain a proper HDF5 addressing ("@..."),
            or verbatim if it's not a string or a slicing tuple.
        '''

        path_mangler = lambda p: p if (p.find('/')>=0 or p[0]!='@')\
            else '@%s/%s' % (self.pathKeys['measurement'], p)
        
        if isinstance(locator, str):
            return path_mangler(locator)
        
        elif isinstance(locator, tuple) and isinstance(locator[0], str):
            return (path_mangler(locator[0]), *(locator[1:]))

        else:
            return locator


    def read (self, frameSlicer, _h5like=None, lean=False, **dataKeys):
        '''
        Returns a bunch of image(s) starting with `start`
        (default: last image taken). If `number` is None,
        all the images to the end of the scan are read.
        Parameters:
        
          - `frameSlicer`: A slice object for selecting frames within a scan
            series.

          - `roiSlicer`: Additional slicer or tuple of slicers to apply to the
            a data frame, e.g. for selecting only specific parts of the data
            frame (region-of-interest). Note that this slicer will be applied
            to *all* data keys to retrieve (see also the `dataKeys` parameter
            below), to the extent to which they have the correct dimensionality.

          - `_h5like`: A h5py-node-like object to use for reading. If none is
            specified, the one generated by the internal `dataSource()[scan]`
            factory is used. Meant only for internal use (e.g. by `.streaks()`)
            and may disappear in the future, but is otherwise safe for any kind
            abuse. The reasoning is that sometimes fake h5py-like objects
            are expensive and open (more so than HDF5 objects), so re-use might
            improve speed.

          - `lean`: If `True`, the "required datasets" (i.e. those that were
            specified with `__init__()`) are omitted from the result. The
            default is `False`.
        
          - `**dataKeys`: Data sets to retrieve; the name of the parameter will
            result in a key, and the value of the parameter should refer to the
            corresponding HDF5-like dataset path. This is passed to
            `HdfCarver`, so please also refer to the documentation
            of that class.
        
        '''        
        data = {} if lean else self.requiredDatasets.copy()
        data.update({k:self.__defPath(v) for k,v in dataKeys.items()})
        
        carver = HdfCarver(data, paths=self.pathKeys)
        
        if _h5like:
            return carver(_h5like, slicer=frameSlicer)
        else:
            with self.dataSource() as h5:
                return carver(h5[self.h5scan], slicer=frameSlicer)
    

    def streaks(self, slicer_generator=None, by=None, _h5like=None, retry=0, **dataKeys):
        '''Uses `read()` to return chunks of the data as specified by `slicer_generator`.
        
        The `addrproc` and `sliceproc` parameters are passed on to `read()`.
        
        Args:
            slicer_generator: enumerable (array, list, generator function...) which
              creates `slice()` objects, one for every streak of the family.
              See `nx5d.streaks` for a list of easily available slicer generators.
              Tip: `numpy.s_` can be used here for one-off streaks.
              This defaults to `slice(None)`, which returns all the data in one
              swoop. If this is a callable, it is interpreted as a factory function
              for a slicer generator, i.e. the actual generator is created calling
              this function with the h5py-like node as its sole parameter.
        
            _h5like: h5py-like object to use for reading. If none is specified
              (default), then the internal `dataSource()` method is being used,
              which initializes a new h5py object as specified in `__init__()`.
        
            retry: If non-zero or not `None`, file opening/reading
              is retried on `OSError` type of exceptions up to `retry`
              number of seconds (fractional number also possible)

            **dataKeys: labels (as keys) and HDF5-paths compatible with
              `HdfCarver` (as values) for the data we're interested
              in.

        Returns: this is a generator function. Each generator invocation yields
          a dictionary with the keys from `dataKeys`, and values loaded
          from the HDF5-like object, conforming to the corresponding slice
          object from `slicer_generator`.
        '''

        with _h5like or self.dataSource() as h5:
            scan = h5[self.h5scan]

            # Courtesy to the user: if the slicer generator is a callable,
            # treat it, in fact, as a factory (for a slicer generator).
            slc_gen = slicer_generator(scan) if hasattr(slicer_generator, "__call__") \
                else slicer_generator
            
            for slicer in (slc_gen or (slice(None),)):
                begin = time.time()
                while True:
                    now = time.time()
                    try:
                        yield self.read(slicer, **dataKeys, h5like=scan, lean=False)
                        break

                    except OSError:
                        if not retry or (now-begin) > retry:
                            logging.error("Final error, %f seconds after initial start" % (now-begin))
                            raise
                        else:
                            logging.debug("OSError, retrying for another %f seconds" % (now-begin-retry))

                    except DatasetEmpty:
                        logging.debug("End of dataset")
                        return

    
    def numFrames(self, dataset, _h5like=None):
        '''
        Reads (returns) the number of frames from the HDF5 file
        for the dataset refered to by `dataset`.
        '''
        
        carver = HdfCarver({'data': self.__defPath(dataset)}, paths=self.pathKeys)

        if _h5like:
            return carver(_h5like, nodeOnly=True)['data'].shape[0]
        else:
            with self.dataSource() as h5:
                return carver(h5[self.h5scan], nodeOnly=True)['data'].shape[0]



#!/usr/bin/python3

from pandas import MultiIndex
from xarray import Dataset
from dask import array as da

'''
Experiment geometry defintion dictionary as needed by `LazyQMap`.
Strongly influenced by typical `xrayutilities` API. This was 
the correct geometry at one point in time (for proposal 3381/Bargheer
in April 2023), but might have changed for *your* experiment.
'''
ExperimentTemplate = {
    "goniometerAxes": {'omega': 'y-', 'chi': 'z+', 'phi': 'x+'},
    "detectorTARAxes": {'twotheta': "y-", 'a': None, 'r': None},

    "imageAxes": ("y-", "x+"),
    "imageSize": "$/detector_size",
    "imageCenter": "$/detector_centre",

    # same unit as imageChannelSize
    "imageDistance": "@/sdd",

    # same unit as imageDistance (mm)    
    "imageChannelSize": ("@/pixel_size", "@/pixel_size"),

    "imageSize": "@/detector_size",

    'sampleFaceUp': 'x-',

    'beamDirection': (0, 0, 1),
    'sampleNormal': (1, 0, 0),

    "beamEnergy": "@/photon_energy",
}

'''
Template for loading of angle data from DAMNIT HDF5 files
'''
DmnAnglesTemplate = {
    'goniometerAngles': {'omega': '@/omega', 'chi': '@/chi', 'phi': '@/phi'},
    'detectorTARAngles': {'twotheta': '@/twotheta', }
}

'''
Template for loading if detector (image) data form DAMNIT HDF5 files.
ACHTUNG, these are typically still giant amounts of data (10k+ images).
Loading these as a numpy array from HDF5 is a BadIdea(tm). Typically,
you'd want to reference them as Dask arrays and apply some further
reduction before holding them in memory.

See `dmn_load_dataset()` as an entry point to a better workflow.
'''
DmnImagesTemplate = {
    'images': "@/images"
}

def dmn_load_dataset(h5like, *small_keys, index_keys=None, **large_keys):
    ''' Loads a DAMNIT-like HDF5 file into an xarray dataset.
    
    DAMNIT-like HDF5 files have a flat structure with (essentially) flat structure,
    where each data set has the same length N and data only differs in additional
    dimensions (i.e. angular data and/or other metadata being essentially
    1-dimensional arrays of length N, while detector images are huge 3D datasets
    with NxWxH items).
    
    The first dimention is either indexed by a numerical sequence (defaults to
    `range(N)`), or by more complex `pandas.MultiIndex` structures. If it needs
    to be a multi-index, then the data keys which make up the multi-index
    can be specified.
    
    The `small_keys`, i.e. 1-dimensional data, is typically referenced directly,
    to be fully loaded into memory.
    
    The `large_keys`, i.e. multi-dimensional data which potentially busts memory
    limits, is loaded as `dask` arrays, and needs to be specified with either a
    3-tuple reprenting the Dask chunking, or at least an integer parameter
    specifying the chunking along the 1st ("index") axis. If it's the latter,
    chunking along the other two dimensions is taken from the dataset's `.shape`
    (i.e. no chuking is performed along higher dimensions).
    
    Args:
    
        h5like: HDF5-like object to read data from; it's expected to be either
          a `h5py.File()` object or a node.
        
        *small_keys: each additional positonal parameter is expected to be a
          string, denoting a dataset to load the "default way" (i.e. will
          ultimately end up as an numpy array, fully into RAM).
        
        index_keys: list or tuple of strings which make up the first dimension
          index. Default (`None`) will lead to creating of an integer counter
          as index.
          
        dmetric_keys: list with detector metric datasets (obsolete?)
            
        **large_keys: each additional named parameter is interpreted to designate
          a dataset (as key) which is to be referenced lazily (e.g. as a Dask
          array). The value represents chunking -- a single integer uses that
          chunk size in the first dimension and reuses the full length in all other
          dimensions, and a tuple is used as a full chuking specification for
          `dask.array.from_array(..., chunks=...)`.
          
    Returns: an `xarray.Dataset` with the requested data.
    '''
    
    if index_keys is not None:
        index = MultiIndex.from_arrays([h5like[k] for k in index_keys],
                                       names=tuple([k for k in index_keys]))
        
    else:
        index = range()
            
    data_vars = {}
    coords = {}
    
    # This is the data we'll reference directly in the h5like dataset.
    # It'll eventually end up as numpy arrays (i.e. fully in memory)
    for k in small_keys:
        data = h5like[k]
        sh = h5like[k].shape
        data_coords = ['index'] + [f"{k}_{i}" for i in range(1,len(sh))]
        data_vars[k] = (data_coords, data)
        if index is None:
            index = range(sh[0])
        
    # This is the data we'll reference only as dask arrays.
    # The value to the keys are chunking information.
    for k,chk in large_keys.items():
        sh = h5like[k].shape
        data_coords = ['index'] + [f"{k}_{i}" for i in range(1,len(sh))]
        if hasattr(chk, "__getitem__"):
            chunks = chk
        else:
            # take chk as chunking along the 1st dimension, and keep all
            # the other dimensions as they are
            chunks = tuple([int(chk)] + [*sh[1:]])
        data = da.from_array(h5like[k], chunks=chunks)
        data_vars[k] = (data_coords, data)
        if index is None:
            index = range(sh[0])
    
    if index is None:
        return None
        
    coords['index'] = index
    
    return Dataset(data_vars=data_vars, coords=coords)



#!/usr/bin/python3

from xarray import Dataset, DataArray
from numpy import array as np_array
from nx5d.h5like.tools import HdfCarver

""" Contains various types of slicer generators to use with the `.streaks()` method
of `nx5d.scan.ScanReader`.
"""

__all__ = [ "s_length", "s_delta" ]

def s_length(val, offset=0, end=None):
    '''Generator for `slice()` objects which return sequential data pieces of lenth `val`,
    i.e. essentially `slice(0,val), slice(val,2*val), ...`.

    Since we don't know how long the full dataset will be, this continues forever,
    unless the parameter `end` is set to a maximum value. The assumption is that
    higher application layers will know when to stop retrieving data.
    The `ScanReader`, for instance, will stop once the `EmptyDataset` exception
    is encountered (...which is the case when a HDF5-like data container returns
    a dataset with length 0 in the first dimension).
    '''
    while (end is None) or (offset<end):
        yield slice(offset, offset+val, None)
        offset += val


def s_delta(val, offset=0):
    '''Generator for `slice()` objects with specified stride,
    i.e. `slice(0,None,val), slice(val,None,val), ...`.
    '''
    for i in range(0, val):
        yield slice(offset+i, None, val)


class Pabst:
    ''' The "Patten-based streaker" builds streaks by evaluating patterns in HDF5 data.

    "Streaking" is the act of repeatedly applying a parametrized set of filters
    on data, typically until *all* data has been covered once. For instance:
    an `s_delta` streak applies the "give me every N-th dataset" filter repeatedly --
    on first run starting with the 0th (then 0+Nth, then 0+2Nth, ...), then starting
    with the 1st (then 1+Nth, then 1+2Nt, ...).
    In simple cases this can be done without knowledge of that the main data itself is
    (e.g. detector images), or parameters associated with the main data (e.g. angles
    or temperatures). But a lot more powerful streaker generators can be built if
    slicing decisions can me made on the basis of data contents.

    This what `Pabst` does, and the main mechanism to achieve that is recursive
    refinement of slicing parameters. This is a two-step process: first, parameters
    relevant for the *definition* of the streaks must be read from the data source;
    second, the slicing objects must be built and presented in a way as to be
    used in `.streaks()` of `nx5d.scan.ScanReader`.
    '''

    def __init__(self, h5like=None, **params):
        '''Initializes the recursive refinement streaker with necessary parameter data.

        Args:
            h5like: A HDF5-like object, or a factory which creates one. If this is
              a callable, it is interpreted as the latter.

            **params: Each parameter name is a key by which subsequent calls
              to refinement methods can refer to data. The contents of the
              parameters need to be addresses usable by `HdfCarver`.
        '''


        # FIXME: Pabst is going to be used something like this:
        #
        #     reader.streaks(slicer_gen=Pabst(...).set(0,1).length(20), ...)
        #
        # Now inside .streaks(), two things could happen. Either using
        # the slicer generator directly:
        #
        #     with dataSource() as h5:
        #         for s in slicer_generator: ...
        #
        # Or checking whether the slicer generator is actually a callable,
        # and use that as a "factory" for a generator which also receives
        # a h5like as a parameter:
        #
        #     with dataSource() as h5:
        #
        #        sgen = slicer_gen(h5) \
        #                  if hasattr(slicer_gen, "__call__") \
        #                  else slicer_gen
        #
        #        for s in sgen: ...
        #
        # In the first case, the h5like would have been already created/used
        # *outside* of the .streaks() function, namely when creating the Pabst
        # (i.e. something like: .streaks(Pabst(h5like, ...), ...). And then
        # presumably the same h5like would've been opened later again, *within*
        # the .streaks() call.
        #
        # The latter has the advantage that we need to open the H5-like (which is
        # sometimes expensive, in particular when using fake h5py APIs) only once,
        # within the .streaks() function, and we get to reuse it. But this would
        # mean that Pabst must not use it here, in the __init__(), and must
        # actually reuse it only within the __call__().
        #
        # NOTE: also, note that for the true HDF5 format, the official HDF5 reader
        # implementation (which is also used by h5py) has a botched up memory
        # management which relies on a *global* lock -- yes, "global" as in
        # "library wide", even when acting on *different* HDF5 files. This means
        # that for very large datasets, it is actually less desireable to
        # stremline opening of files, and that more desireable to repeatedly
        # re-open files (even if it's the same file) over and over again, if this
        # can be done from different processes (as different processes, being in
        # different memory namespaces, will circumvent the library-wide locking).
        # However, intelligently making use of this feature actually requires
        # employing a multi-processing toolkit within Python, e.g. "multiprocess",
        # and this is something which is going to require extra work and
        # consideration in using with .streaks() anyway -- something we don't do
        # yet.
        #
        # So we will go the "open HDF5 as little as possible" approach for now,
        # which means we'll store the h5like and parameter template here, and
        # actually only use them on __call__().
        #

        self._h5like = h5like
        self._params_template = params
        self._refine = []


    def moo(self, **kwargs):
        self._refine.append(kwargs)
        return self


    def __call__(self, h5like=None):
        ''' Triggers the refinement, optionally using the specified `h5like`.
        '''

        # We prefer the supplied one, but will gladly take the stored one
        # (i.e. supplied with __init__) if none is supplied here.
        h5obj = h5like or self._h5like
        if hasattr(h5obj, "__call__"):
            with h5like() as h5:
                params = HdfCarver(self._params_template)(h5)
        else:
            params = HdfCarver(self._params_template)(h5obj)

        # At this point, self.params hs a dict() containing data.
        # What we need to do now is loop through the data items
        # and yield the slices -- one at a time. What we need to
        # to *before* that is apply the refiners. We like to work
        # on an xarray.Dataset().
        tmp = next(iter(params.items()))
        xdata = Dataset({'INDEX': np_array(range(tmp[1].shape[0]))})
        for k,p in params.items():
            dims = ["INDEX"] + ["%s_%d" % (k,i) for i in range(1,len(p.shape))]
            xdata[k] = (dims, p)
        
        return xdata

        #print("xdata coords:", xdata.coords)
        #print("xdata:", xdata)

        #print("idx 3:", xdata.sel(idx=3))

        # xdata now has all the relevant data.
        # (hopefully) self._refine is a list of dictionaries (key, value)
        # to refine the index list with. Iterate through it.
        #result = xdata
        #for ref in self._refine:
        #    print("refine:", ref)
        #    #result = result.sel(**ref)

        #return result



#!/usr/bin/python3

import xrayutilities as xu
from xarray import DataArray, Dataset
from xarray import concat as xr_concat
import numpy as np

import logging

import time

import dask

from tqdm import tqdm

'''
Example for device geometry:

```
{
  "goniometerAxes": ('y+', 'z+', 'x+'),

  "detectorTARAxes": (None, "z+", None),
  "imageAxes": ("x+", "z+"),
  "imageSize": (1024, 768),
  "imageCenter": (0.5, 0.5),
  "imageChannelSpan": (None, None),
  "imageChannelSize": (None, None),

  "sampleFaceUp": 'x+',
  "beamDirection": (0, 1, 0),
}

```
'''

class InsufficientData(RuntimeError):
    # raised when data sets are missing
    pass

class InsufficientAngles(InsufficientData):
    # specifically for angle datasets
    pass

class UnsuitableData(RuntimeError):
    # raised when data format is bad
    pass
    
class QMapper:
    ''' Base class for all Xrayutilities based Q-space mapper.

    This essentially just initializes the experimental setup
    for xrayutilities and prepares for accepting "useful" data.
    The relevant Xrayutilities algorithm wrappers are also
    implemented here.

    This is a fully functional implementation with a clean
    separation between static setup data (experiment geometry)
    and measurement data (angles and images). Subclasses may still
    implement more specific usage APIs.

    Example
    ```

    # Defining an experimental geometry. We put this in a dictionary
    # for clarity, but we could just as well pass the parameters to
    # QMapper(...) below one by one.
    exp_setup = {
        'beamDirection': (0, 1, 0),
        'beamEnergy': 9600.0,
        'imageAxes': ('x-', 'z-'),
        'imageCenter': (90, 245),
        'imageChannelSize': (0.172, 0.172),
        'imageDistance': 720.0,
        'imageSize': (195, 487),
        'sampleFaceUp': 'z+',
        'sampleNormal': (0, 0, 1)

        # keys here will be used to find the angles in the xarray.Dataset
        'goniometerAxes': {
            'phi': 'x+',
            'chi': 'y+',
            'omega': 'z+'
        },

        # These last two keys of the detector axes will be ignored,
        # because angle spec is None. Yet we still need to define them
        # even if the detector doesn't have an Azimuth or Rotation axis.
        'detectorTARAxes': {
            'twotheta': 'x+',
            'a': None,
            'r': None
        },
    }

    # We demonstrate the layout of an xarray Dataset suitable for Q-mapping.
    # In a real example, instead of generating the data, we'd read it
    # (e.g. from a HDF5 file)
    raw_data = xarray.Dataset(
        data_vars={
          'chi':      ('index', np.array(...)),
          'phi':      ('index', np.array(...)),
          'omega':    ('index', np.array(...)),
          'twotheta': ('index', np.array(...)),
          'pilatus':  (('index', 'x', 'y'), np.array(...)),
        },
        coords={
          'index': np.array(range(N)), # N is the number of images/angles here
          'x': np.array(range(195)),   # 195 is the image width, see exp_setup above
          'y': np.array(range(487))   # 487 is the image height in pixels
        })

    # Defining the mapper
    mapper = QMapper(**exp_setup)

    # This is the most simple approach: call .qmap() and let it figure out
    # useful defaults (which it will, given the Dataset above).
    q_data = mapper.qmap(raw_data)

    # Different approach: specify explicitly which image to transform...
    q_data = mapper.qmap(raw_data, images="pilatus")

    # ...or which angle sets to use...
    q_data = mapper.qmap(raw_data, angles=("chi", "phi", "omega", "twotheta"))

    # ...or both.
    q_data = mapper.qmap(raw_data, angles=("chi", "phi", "omega", "twotheta"),
                         images="pilatus")

    # Control Q-space grid size
    q_data = mapper.qmap(raw_data, qsize=(100, 100, 200))

    # Or reduce the number of dimensions (i.e. 2D Q-space map)
    q_data = mapper.qmap(raw_data, dims=("qx", "qz"))

    # Or both at the same time (using a dictionary for qsize instead of a tuple)
    q_data = mapper.qmap(raw_data, qsize={"qx": 100, "qz": 200})
    ```

    This will result e.g. in transforming a raw_data dataset like this:
    ```
    >>> raw_data
    <xarray.Dataset>
    Dimensions:   (index: 64, x: 195, y: 487)
    Coordinates:
      * index     (index) int64 0 1 2 3 4 5 6 7 8 9 ... 55 56 57 58 59 60 61 62 63
      * x         (x) int64 0 1 2 3 4 5 6 7 8 ... 187 188 189 190 191 192 193 194
      * y         (y) int64 0 1 2 3 4 5 6 7 8 ... 479 480 481 482 483 484 485 486
    Data variables:
        phi       (index) float64 0.0 0.0 0.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 0.0 0.0
        chi       (index) float64 0.0 0.0 0.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 0.0 0.0
        theta     (index) float64 12.0 12.01 12.02 12.03 ... 12.6 12.61 12.62 12.63
        twotheta  (index) float64 24.0 24.01 24.02 24.03 ... 24.6 24.61 24.62 24.63
        pilatus   (index, x, y) float64 0.0 0.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 0.0 0.0
    ```

    into something like this (in the default case):
    ```
    >>> q_data
    <xarray.DataArray (qy: 64, qx: 195, qz: 487)>
    array([[[0., 0., 0., ..., 0., 0., 0.],
            [0., 0., 0., ..., 0., 0., 0.],
            [0., 0., 0., ..., 0., 0., 0.],
    ...
            [0., 0., 0., ..., 0., 0., 0.],
            [0., 0., 0., ..., 0., 0., 0.],
            [0., 0., 0., ..., 0., 0., 0.]]])
    Coordinates:
      * qy       (qy) float64 -0.5787 -0.574 -0.5693 ... -0.293 -0.2883 -0.2836
      * qx       (qx) float64 -0.2423 -0.2409 -0.2395 ... 0.0337 0.03513 0.03657
      * qz       (qz) float64 1.72 1.721 1.722 1.723 ... 2.279 2.28 2.281 2.282
    ```
    '''

    def __init__(self, **experiment_setup):

        ''' Initializes the experimental geometry.

        The arguments to `.__init__()` are inspired from the xrayutitlities
        `Experiment` classes, but they're intended to be fairly generic
        and represent the actual physics, not a specific coded implementation
        of it. They should be the same for all backends, even if different from
        xrayutilities.
        
        Args:
            **experiment_setup: parameters for the experimental setup. Refer
              to `.init_experiment()` for a complete documentation of the
              parameters.

        '''

        self.init_experiment(**experiment_setup)
        self.Ang2Q = self.hxrd.Ang2Q

        # additional gridder / ang2q setup variables
        self.gridderDict = {}
        self.ang2qDict = {}
        

    def init_experiment(self,
                        beamEnergy=None,
                        beamDirection=None,
                        detectorTARAxes=None,
                        goniometerAxes=None,
                        imageAxes=None,
                        imageCenter=None,
                        imageChannelSize=None,
                        imageChannelSpan=None,
                        imageDistance=None,
                        imageSize=None,
                        sampleFaceUp=None,
                        sampleNormal=None,
                        roi=None):
        
        '''Initializes the experiment setup representation.

        This is typically an xrayutilities HXRD object or similar,
        with specified device and sample geometry. The optional parameter `roi`
        restricts angle-to-Q conversion to solely this region, if it is
        specified. This is a good way to save significant amounts of computing
        time.

        Args:
        
            beamEnergy: the energy of the incoming X-ray beam, in eV.

            beamDirection: direction of the X-ray beam. 
            
            detectorTARAxes: the direction of each of the 3 possible movement
              axes of the detector (tilt, azimuth, rotation). Note that there are
              always 3 of these axes each with a very specific purpose in
              `xrayutilties`. If your detector lacks any, complement with `None`.
              This can either be a simple enumerable (list or tuple), to specify
              only the axis orientations; or this can be a dictionary, to *also*
              specify the names in addition to the orientation.
        
            goniometerAxes: the direction of each of the goniometer axes in
              the `[xyz][+-]` notation. This is a variable-sized array, as there
              can be several axes in any goniometer, and `xrayutilities` apparently
              magically knows what to do.               
              Unlike the detector angles, this can hold an arbitrary number
              of angles, the only restriction (according to `xrayutilities` docs)
              being that these must be in from the outer to the inner rotations.
              This can either be a simple enumerable (list or tuple), to specify
              only the axis orientations; or this can be a dictionary, to *also*
              specify the names in addition to the orientation.
        
            imageAxes: the direction of the image axes (x and y) at zero angles.
              The positive direction of the axes should coincide with increasing pixel
              index in the data.
        
            imageCenter: this is the position of the center pixel, either absolute
              (integer pixel numbers), or relative to the sensor size (as specified in
              `imageAxes`). If the number is in the range 0.0..1.0, then relative
              positioning is assumed.
        
            imageChannelSize: for Q transformation, ultimately the relation between
              every specific on the detectors and the angle
              of the incoming beam activating that specific pixel is needed. There
              are two distinct ways of specifying this: either using the "channel span",
              i.e. the size, in degrees, of each pixel, in horizontal/vertical direction,
              or by a distance parameter (from detector to sample) and a pixel size.
        
              `imageChannelSpan` is either a single number or a 2-tuple specifying
              how many degrees one channel takes. `imageChannelSize` specifies the
              spatial size of a pixel relative to the distance between the sample
              and the sensor.

            imageChannelSpan: (overrides `setup["imageChannelSpan"]`),
              see `imgeChannelSize`.
        
            imageDistance: (overrides `setup["imageDistance"]`),
              distance of the detector from the center of rotation.
        
            imageSize: (overrides `setup["imageSize"]`), width and height of one detector
              image.
        
            sampleFaceUp: (overrides `setup["sampleFaceUp"]`),
              direction of the "sample surface facing up", a.k.a.
              "sampleor" (sample orientation) in `xrayutilities` lingo.
              This is the orientation of
              the sample surface at zero angles. This is either an axis notation
              (`[xyz][+-]`) or one of the special words `det`, respectively `sam`.
        
            sampleNormal: (overrides `setup["sampleNormal"]`), not sure
              what this is... in `xrayutilities`.

        Returns: the internal representation of the experiment (typically an
            xrayutilities.Experiment instance, when xrayutilities is used as a backend).
            The internal representation is also stored within the object.
        '''
        
        def __2tuple(data, name=""):
            ## Returns a 2-tuple (X, Y) from a variety of data sets:
            ## - from a 2-tuple :-) or an array with size 2
            ## - from an array (N, 2) (returns the tuple of the first elements)
            ## - from an array or tuple (2, N) (returns the first element)
            if not hasattr(data, "__len__"):
                raise UnsuitableData(f"{name}: needs to be a 2-tuple")
            if len(data) == 2:
                if hasattr(data[0], "__len__"):
                    return ([data[0][0], data[1][0]])
                else:
                    return ([data[0], data[1]])
            if len(data) > 2:
                return data[0]
            raise UnsuitableError(f"Oops: don't know how to handle {name}: {data}")


        ## Axes can be either dict() or enumerables; try to sort them out.
        if hasattr(goniometerAxes, "keys"):
            gonAxes = [goniometerAxes[g] for g in goniometerAxes]
            setupAxesKeys = [k for k in goniometerAxes]
        else:
            gonAxes = goniometerAxes
            setupAxesKeys = []

        if hasattr(detectorTARAxes, "keys"):
            detAxes = [x[1] for x in filter(lambda x: x[1] is not None, detectorTARAxes.items())]
            setupAxesKeys += [x[0] for x in filter(lambda x: x[1] is not None, detectorTARAxes.items())]
        else:
            detAxes = [x for x in filter(lambda x: x is not None, detectorTARAxes)]
            setupAxesKeys = None ## don't store axes keys if we don't have a complete set

        # beamEnergy is usually supposed to be a scalar, but sometimes an array
        # (one for each image) will be supplied. In that case, retrieve only the first.
        beamEnergy = beamEnergy if not hasattr(beamEnergy, "__len__") else beamEnergy[0]

        qconv = xu.experiment.QConversion(sampleAxis=gonAxes,
                                          detectorAxis=detAxes,
                                          r_i=beamDirection,
                                          en=beamEnergy)

        self.hxrd = xu.HXRD(idir=beamDirection,
                            ndir=sampleNormal,
                            sampleor=sampleFaceUp,
                            qconv=qconv,
                            # Workaround for buggy xrayutilities: repeat the beam energy
                            en=beamEnergy)

        imageCenter = __2tuple(imageCenter, 'imageCenter')
        imageSize = __2tuple(imageSize, 'imageSize')
        imageDistance = imageDistance[0] if hasattr(imageDistance, "__len__") \
            else imageDistance

        logging.debug("image distance %r, size %r, center at %r" % \
                      (imageDistance, imageSize, imageCenter))
        
        if imageCenter[0] <= 1 and imageCenter[1] <= 1:
            # It's a floaring-point number, relative to the detector size

            # similar considerations for imageSize as for imageCenter: expected to be
            # a 1D array with length 2, but will also accept a 2D array
            # with shape (N, 2).
            
            # FIXME: really, REALLY need to fix this. This is really ugly.
            imgs = imageSize
            assert imgs is not None
            assert imgs[0] is not None
            assert imgs[1] is not None            
            imageCenter = tuple([c*s for c,s in zip(imageCenter, imgs)])

        chSizeParm = {}
        if imageChannelSpan is not None:
            # channelSpan is degrees/channel, but need to pass channels/degree to Ang2Q
            imageChannelSize = self.__2tuple(imageChannelSpan, 'imageChannelSpan')
            chSizeParm = {'chpdeg1': 1.0/imageChannelSpan[0],
                          'chpdeg2': 1.0/imageChannelSpan[1] }

        elif imageChannelSize is not None:
            # Ang2Q takes one explicit distance parameter, but we're assuming that
            # channelSize is relative to the distance itself (putting the distance
            # always at 1.0 units)
            imageChannelSize = __2tuple(imageChannelSize, 'imageChannelSize')
            logging.debug("pixel size: %r" % (imageChannelSize,))
            chSizeParm = { 'pwidth1':  imageChannelSize[0],
                           'pwidth2':  imageChannelSize[1],
                           'distance': imageDistance }

        else:
            raise RuntimeError("Experiment setup needs either the "
                               "channel span or channel size")

        if roi is None:
            roi = (0, imageSize[0], 0, imageSize[1])
            
        self.hxrd.Ang2Q.init_area(detectorDir1=imageAxes[0],
                             detectorDir2=imageAxes[1],
                             cch1=imageCenter[0],
                             cch2=imageCenter[1],
                             Nch1=imageSize[0],
                             Nch2=imageSize[1],
                             tiltazimuth=0,
                             tilt=0,
                             detrot=0,
                             roi=roi,
                             **chSizeParm)

        ## save some relevant setup parameters for later use
        ## (e.g. for auto-detecting datasets in .qmap())
        self.setupRoi = roi
        self.setupImageSize = imageSize
        self.setupAxesKeys = setupAxesKeys
        self.setupImageAxes = imageAxes

        logging.debug(f'Remembering setup axes keys: {self.setupAxesKeys}')
        
        return self.hxrd


    def setupGridder(self, **gridderDict):
        self.gridderDict = {}        
        self.gridderDict.update(griderDict)


    def setupAng2Q(self, **ang2qDict):
        self.ang2qDict = {}
        self.ang2qDict.update(ang2qDict)
    

    def qmap(self,
             xdata,
             images=None,
             angles=None,
             qsize=None,
             dims=None,
             retain=True,
             output=None):
        ''' Performs Q-space mapping on `xdata`.

        The detector image data to map into Q-space (see `images` and `xdata`
        parameters) must have the image same size as the `imageSize` parameter
        that was passed to `.__init__()` of no region-of-interest (`roi`)
        was specified. If a `roi` was specified to `.__init__()`, then
        the image data must be the size of the `roi`, and the corresponding
        pixel index must start with the lower numbers of the corresponding
        `roi` dimension.

        Args:
        
            xdata: `xarray.Dataset` with detector data to be transformed
              and angles.

              By default, the name of the detector data (array)
              within the set is extracted from the `images` parameter,
              or is guessed automatically by comparing the 2nd and 3rd
              dimension with `imageSize`. If multiple matching detector
              image sets are found, the first that matches is used.

              The name of the angles are used by reading the `goniometerAxes`
              and `detectorTARAxes` from `.__init__()`, but can be overridden
              by the `angles` parameter.

            images: a string representing name of the detector data to transform
              within the `xdata` dataset.

            angles: a tuple of strings representing names of the angle datasets
              within `xdata`. If this is specified, it must first list all goniometer
              angles (from outer to inner), then all detector axes, as a flat
              names tuple (i.e. non-nested). All the angles datasets must have the
              same first dimension as the images dataset.

            qsize: Can be either a tuple `(w, h)`, or `(w, h, d)` of the resulting Q
              image, or a dictionary `{'qx': w, 'qy:..., ...}`.
              If it is `None`, the size of the original angular image dimension(s)
              is used for a 3D Q-mapping.

            dims: List with dimension names "qx", "qy" or "qz", or any combintion
              thereof, for resulting Q image. This effectively dictates what
              kind of Q-mapping is performed (1D for a single dimension name,
              2D if two names are specified, or 3D of all 3 are specified), and which
              projection (i.e. `("qx", "qy")`, or `("qz", "qx")`, ...).
              The spatial designations (x, y or z) are in consistency with the
              corresponding axis definitions of the `.__ini__()` parameters, where
              each reciprocal space (Q-) coordinate is correctly defined as perpendicular
              on the other two real space coordinates in `.__init__()`.
              If this is `None` (the default value), the keys of `qsize` are used;
              if those are *also* `None` (also the default value), then a 3D Q-space
              mapping is performed.
              
            retain: if True, extra data variables and coordinates
              from `xdata`, which are not being processed during the Q-mapping (i.e.
              everything else besides image data and angles), is being transfered
              over to the output data. This effectively retains per-transformation
              metadata. The default is "auto", which results to `False` here,
              but `True` in `.qmap_groupby()`.
              
            output: Name of the output array in the resulting dataset. If `None`,
              it will be the same as the name of the input data.
        
        Returns: an `xarray.Dataset` (or `DataArray`?) with the designated detector
          data (see `images`) on a corresponding reciprocal space grid (see `qsize`),
          and their corresponding Q-axes values.
        '''

        image_name  = self._select_image_name(xdata, images)
        angle_names = self._select_angle_names(xdata, angles)        
        grid = self._select_grid_size(qsize, dims, xdata[image_name].shape)        

        return self._invoke_area_qconv(xdata, image_name, angle_names, grid,
                                       retain=retain in (True, 'auto'),
                                       output=output if output is not None else image_name)
    
    
    def qmap_single(self, xdata, *args, **kwargs):
        ''' Convenient wrapper for `.qmap()` to use when data only has a single image.
        
        "Regular" `.qmap()` will choke, as it is expecting a batch of angular pixel
        data, and a batch of angles. On single images, there's only one of each
        (each angle, image etc), and there's no 3rd dimension.
        
        The easy thing to do is simply extend `xdata` by an extra dimension. However,
        one thing we *also* want to do is modify implicit settings for `dims` to
        produce a 2D Q-map (since that's what the user will be expecting).
        
        NOTE: try using `.groubpy(squeeze=False)` instead.
        '''
        default_dims = tuple([f'q{x[0]}' for x in self.setupImageAxes])
        qdata = self.qmap(xdata.expand_dims('index'), *args, **kwargs)
        return qdata
    
    
    def qmap_groupby(self,
                     data,
                     groupby=None,
                     combine="qcoord",
                     qquant=1e6,
                     *args, **kwargs):
        ''' Invokes the Q-space mapping on every member of a `.groupby()` result.
        
        Think of this as a fancy `xdata.groubpy(...).map(QMapper().qmap())` call.
        The split-process-combine pattern of `.groupby()` is very powerful and
        allows to write expressive data analysis code -- which we want. However,
        combining results of `.qmap()` usually lead to unintended results, simply
        because every single call comes with its own set of qx, qy and qz
        coordinates.
        
        Mostly, these are "the same" in the physical sense, but `xarray` doesn't
        recognize them as the same because of minor numerical variations. To
        combine them (and still have some failsafe / restored functionality when
        they're actually *not* the same, just similar), we scale and transform the
        coordinate vectors into integers after `.qmap()` invocation, but before
        combining. Then after combination, we re-scale them back to their original
        magnitude and transform them to float.
        
        Args:
            data: `xarray.Dataset` or `DatasetGroupBy`. If it's a dataset, it is
              grouped according to the `groupby` parameter(s). Otherwise the groups
              are used as they are.
              
            groubpy: string or enumerable to pass to `.groupby()`. Only used if
              `data` is not a `DatasetGroupBy`.
              
            combine: how to combine data after having transformed each group item
              through `.qmap()`. The following options are accepted:
                - `None` or "*none*" does not recombine, just returns
                  a list of result objects.
                - "*qcoord*" assumes that all Q-coordinates, to all of the transformed
                  group items, are the same, and variations are just owing to numerical
                  differences. Recombination is therefore performed explicitly by
                  `.qmap_groupby()` by taking the first set of Q-coords, and
                  overriding/reusing it for every other group item.
                - "*groupby*" goes with whatever the `.groubpy()` function does.
                  It only works when the `groubpy` parameter is not `None`.
                - "*qquant*" is like "groupby", but quantizes the Q-coordinates
                  first, by multiplying them with a factor (specified in `qquant`)
                  and transforming to `int` after Q-mapping, as part of
                  the `.map(...)` work function. The transformation is reversed
                  after grouping, such that the Q-coordinates have (again) the
                  same magnitude and a `float` array type -- but likely less
                  precision.
                  
              qquant: multiplication factor (i.e. magnitude) for the Q-coordinates,
                if `combine` was set to "qquant".
                
              *args: passed to `.qmap()`.
              
              **kwargs: passed to `.qmap()`. If the `retain` argument is set to "auto",
                it is modified to `True`.
        
        '''
        
        if kwargs.get('retain', 'auto') == 'auto':
            kwargs['retain'] = "data_vars"
        
        ## Step 1: splitting (if necessary)
        if hasattr(data, "groups"):
            xgroups = data
        else:
            is_dask = np.array([(data[k].chunks is not None) for k in data]).any()
            input = data.compute() if is_dask else data
            if groupby is None:
                raise RuntimeError(f'Input data is {type(data)}; if that\'s not a DatasetGroupBy, '
                                   f'you need to specify a groupby=... grouping criterion.')
            xgroups = input.groupby(groupby, squeeze=False)
        
        ## Step 2a: Q-mapping and combining in one go (implicitly via .map())
        if combine in ("groupby", "qquant"):
            scale_q_coords = lambda data, factor, dtype: data.assign_coords({
                dn:(dvar*factor).astype(dtype) for dn,dvar in \
                filter(lambda x: x[0] in ('qx', 'qy', 'qz'), data.coords.items()) })
            invoke_qconv = lambda x: scale_q_coords(self.qmap(x, *args, **kwargs), qquant, int)
            qdata = xgroups.map(invoke_qconv)
            return scale_q_coords(qdata, 1.0/qquant, float)
        
        # Alternatively, Step 2b: first Q-mapping..
        qlist = [self.qmap(l[1], *args, **kwargs) for l in tqdm(xgroups)]
        #qlist = [ t. for t in tmp]
         
        # Step 3: ... then combining.
        if combine in (None, False, "none"):
            return qlist

        if combine in ("qcoord",):
            
            # Reconstructing the main groupby index is a bit tricky if the index
            # was a MultiIndex. We need to actually retrieve the MultiIndex components
            # first, add them to the data, then .set_index() on those components with 
            # a new index name (possibly reuse one, if we have one).
            
            # if groupby was an index name, reuse that; otherwise if it was a tuple of
            # variables, use those, and give the index a new name.
            if isinstance(groupby, str):
                index_ckeys = [k for k in data.indexes[groupby].names]
                index_name = groupby
            else:
                index_name = "index"
                index_ckeys = groupby if groupby is not None \
                              else [k[0] for k in data.groups.items()]
                        
            qdata = xr_concat(qlist, dim=groupby, join='override')
            
            if hasattr(data, "groups"):
                logging.warning(f'"data" was a DatasetGroupBy, cannot reconstruct index')
                return qdata
            
            index_components = { k:(index_name, data[k].data) for k in index_ckeys }
            return qdata.assign(index_components)\
                        .set_index(**{index_name: [k for k in index_components.keys()]})

        
        raise RuntimeError(f'You are not supposed to ever end up here (combine={combine}).')


    def _invoke_area_qconv(self, xdata, image_name, angle_names, grid,
                           ang2qObj=None, qquant=None, retain="data_vars",
                           output=None):
        ''' Actual (internal) invocation of the Q-conversion.

        Parameter sorting & processing happens before this. Here the
        actual breakdown of the xarray data into (for xrayutilities usable)
        numpy arrays happens. For small data this is straight-forward
        by just calling `.values` on the data.
        However, but subclasses might want to reimplement this for more
        sophisticated treatment (e.g. using Dask arrays? threads?...
        split-process-combine patterns?)
        
        This function *can*, but *is not intended* to be called from outside
        QMapper. There are some specific arguments used to enhance "user
        experience", in particular when working with large data / Dask clusters,
        but they require intimate knowledge of QMapper's internals.
        
        Args:
            xdata: the `xarray.Dataset` to work on (must contain images and angles)
            image_name: name of the image data var to transform
            angle_names: list of angle data vars, each for one angle, in the
              correct order (see `xrayutilities` or `.__init__()`)
            grid: dictionary with grid size(s) in Q-space
            ang2qObj: `Ang2Q` object from `xrayutilities`. If not specified,
              `self.Ang2Q` is used (which has been created at `.__init__()`
              or `.init_experiment()` time)
            qquant: factor to use for Q-coordinate quantization. If this
              is different from `None`, then the Q-coordinates are multiplied
              by this factor and cast to integer after Q-mapping, but before
              returning the final result. This feature is used internally in
              `.qmap_groupby()` in order to make direct comparison of Q values
              easier, and thus recombination of `.groubpy()` data pieces.
              Steer away from this if you don't know what you're doing.
            retain: if `True` (default here), all data variables and coordinates
              from `xdata` which weren't used in the transformation are retained
              in the output. 
            output: Which name to choose for the output Q-mapped images.
        '''
        
        tmp_i = xdata[image_name] #.values
        tmp_a = [xdata[a].values for a in angle_names]
        result_name = output if output is not None else image_name
        qdata = self._area_qconv(tmp_i, tmp_a, grid, result_name,
                                 self.gridderDict, self.ang2qDict, self.Ang2Q)
                
        if retain in (None, False, "none"):
            return qdata
        
        ## copy extra data and coordinate dimensions over.
        ## ACHTUNG: disabling this, as this will also copy index coordinates,
        ##          which will massively interfere with .qmap_groupby()'s
        ##          attempts to restore indices. Not sure how to deal with
        ##          this yet. (Maybe we should filter out index+index parts?)
        ##
        ## index = xdata.dims[0]
        ## index_parts = xdata.get_index(index)
        ##   ^^^ like this?
        ##
        if retain in ("all", "coords", True):
            logging.warning(f'retain={retain}, but will cowardly refuse to retain coords')
        #    xcoords = filter(lambda x: (x not in qdata.coords), xdata.coords)
        #    new_coords = { k:xdata.coords[k] for k in xcoords }
        #    qdata = qdata.assign_coords(new_coords)
        
        if retain in ("all", "data_vars", True):
            xvars = filter(lambda x: (x not in qdata.data_vars and\
                                     x not in angle_names and\
                                     x != image_name),
                           xdata.data_vars)
            extra_vars = { k:xdata.data_vars[k] for k in xvars }
            for k in extra_vars.keys():
                qdata[k] = xdata[k]
        
        #print(f"qconv data: {[k for k in qdata.keys()]}, retain: {retain}")
        return qdata
        

    def _verify_data(self, image_data, angle_data):
        # check angles / images for dimension integrity and get loud if they don't match
        for a in angle_data:
            if a.shape[0] != image_data.shape[0]:
                logging.error('One full set of angles is expected for each image, but we got this instead:')
                logging.error(f"Images: {image_data.shape}")
                for a in angle_data:
                    logging.error(f"Angles: {a.shape}")
                raise UnsuitableData('Dimension mismatch on data')
        

    def _select_grid_size(self, qsize, dims, data_shape):
        ## Returns a dictionary with the Q coordinate axes we want to have
        ## (i.e. qx, qy, qz...) and their respective sizes in Q-space.
        ## Defaults to a full 3D Q-map of the same size as the original data.

        # first we need to determine the axis names: self.setupImageAxes defines how
        # the image coordinates Width and Height are named; the rest is the remaining
        # one.
        all_axes = ['x', 'y', 'z']
        tmp = [i[0] for i in self.setupImageAxes]
        for t in tmp:
            assert t in all_axes
        full_axis_set = [f'q{x}' for x in filter(lambda x: x not in tmp, all_axes)] \
            + [f'q{x}' for x in tmp]

        default_qsizes = { q:s for q,s in zip(full_axis_set, data_shape) }

        # Common mistake: when specifying only one axis (1D gridding), sometimes
        # people don't want, or remember, to properly write tuples as ("axis",).
        # The error message of trying to iterate through a string instead of
        # an enumerable is also quite cryptic. We try to catch that and
        # "do the right thing".
        if isinstance(dims, str):
            dims = (dims,)
        
        # names: dims takes precedence; if qsize has names, we use them
        # only for identifying data, if dims is defined; otherwise we
        # use them to collect labels, too.
        # if all fails, we fall back on full_axis_set for names.
        names = dims or ([k for k in qsize.keys()] \
                         if hasattr(qsize, 'keys')
                         else full_axis_set)

        all_q_axes = tuple([f'q{x}' for x in all_axes])
        for n in names:
            if n not in all_q_axes:
                raise RuntimeError(f'Requested Q-axis "{n}" not in accepted pool {all_q_axes}')

        if hasattr(qsize, "keys"):
        # qsize is a dict()-like
            return dict(filter(lambda x: x[0] in names, qsize.items()))
        
        elif hasattr(qsize, "__getitem__"):
            # qsize is a simple tuple/array
            sizes = qsize if qsize is not None else data_shape
            if len(sizes) != len(names):
                raise RuntimeError(f'Requested Q dimension names ({names}) '
                                   f'and sizes ({sizes}) mismatch')
            return { q:s for q,s in zip(names, sizes) }
        
        elif qsize is None:
            return { q:default_qsizes[q] for q in names }

        else:
            raise RuntimeError(f'What to do with qsize={qsize}?')

            


    def _select_angle_names(self, xdata, angles):
        ##
        ## Returns a list of angles, according to `angles` (it not None),
        ## or according to the setup goniometerAxes / detectorTARAxes.
        ##
        if angles is None:
            if self.setupAxesKeys is None:
                raise InsufficientAngles(f'No angle data specified -- you need to either set `angles`,'
                                         f'or modify the axes specifications to contain axis names.')
            angles = self.setupAxesKeys
            
        return angles


    def _select_image_name(self, xdata, images):
        ##
        ## Selects the suitable image data vector from `xdata`. Essentially,
        ## we either use what `images` tells us to, or try to auto-guess from
        ## the detector image geometry+roi of the setup data.
        ##
        ## Returns a 3D image array (xarray?).
        ##
        if images is not None:
            image_data = xdata[images]
            if len(image_data.shape) != 3:
                logging.error(f'Data set "{images}" has wrong dimensionality. Expected: 3D image data, '
                              f'got {image_data.shape} instead. '
                              f'Continuing, but this is most likely not what you want to do.')
            roi_size = (self.setupRoi[1] - self.setupRoi[0],
                        self.setupRoi[3] - self.setupRoi[2])
            if image_data.shape[-2:] != roi_size:
                logging.error(f'Data set {images} has wrong image size (expecting '
                              f'{image_data.shape[-2:]}, got {roi_size}). '
                              f'Continuing, but this might crash.')
                
            return images
        
        else:
            for (img_name, img_data) in xdata.data_vars.items():
                _size = np.array((self.setupRoi[1] - self.setupRoi[0],
                                  self.setupRoi[3] - self.setupRoi[2]))
                if len(img_data.shape) != 3:
                    logging.debug(f"Skipping {img_name}: shape != 3D")
                    continue ## data has wrong dimension
                if (np.array(img_data.shape[1:3]) != _size).all():
                    logging.debug(f"Skipping {img_name}: img_size != {_size}"
                                  f" -> {img_data.shape[1:3]}")
                    continue ## image does not match ROI / setup image size
                logging.debug(f'Auto: detector image data is {img_name}')
                return img_name
            
        raise InsufficientData(f'No suitable detector image found among data '
                               f'vars {[k for k in xdata.data_vars]}')

        
    def _calc_qcoord(self, ang2q, angles, aqdict, grid):
        # calculates a Q-coordinate grid
        qindex = { 'qx': 0, 'qy': 1, 'qz': 2 }
        q = ang2q.area(*angles, **(aqdict or {}))
        return [q[qindex[d]] for d in grid.keys()]

    
    def _calc_qmap(self, images, qcoord, grid, gridderDict):
        # Calls the gridder to calculate the Q-images.
        # Returns the gridder object array (containing Q-data and Q-axes)
 
        # Call scheme of all the xrayutilities Gridders is pretty similar.
        #print("o", end="")
        Gridder = getattr(xu, "FuzzyGridder%dD" % len(grid))
        grd = Gridder(*[g[1] for g in grid.items()])
        grd(*qcoord, images.values, **(gridderDict or {}))
        return grd

    def _retr_qdata(self, gridder):
        return gridder.data
        
    def _retr_qaxis(self, gridder, axis):
        return getattr(gridder, f"{'xyz'[axis]}axis")
        
    def _area_qconv(self,
                    images,
                    angles,
                    gridSize,
                    result_name,
                    _gridderDict=None,
                    _ang2qDict=None,
                    _ang2qObj=None):
        '''
        Front to the ang-to-Q conversion, currently only for area data. Parameters:
        `datasets` is either empty, or a single data label. (No multiple label support
        yet.)
        
        Args:
            images: 3D array of images to transform, 1st dimenson number of images,
              dimnesions 2 and 3 as width/height of images. `QMapper` is Dask-aware,
              so if the `images` array is a Dask array, the most compute-intensive
              step (gridding) will be performed in a `dask.delayed()` wrapper
              and a Dask array will be returned as the Q-mapped images.

            angles: list of all necessary angles, in the correct order (first
              goniometer from outer to inner, then available TAR angles)

            gridSize: Python dictionary with Q-space directions as keys, and number
              of data points in Q-space as values

            `_gridderDict`: If specified, this is a dictionary with extra named
              parameters to be passed on to the gridder. Note that this is not portable,
              only works as long as we're using xrayutilities under the hood.

            _ang2qDict: Extra set of parameters to be passed to the data-specific
              `Ang2Q` function (typically `Ang2Q.area()` for stacks of 2D datasets).
        '''
        
        if len(images.shape) != 3:
            raise RuntimeError(f"Don't know how to transform objects of shape {images.shape}")
        
        # For transforming strings to dimension indices
        #qindex = { 'qx': 0, 'qy': 1, 'qz': 2 }

        # Call scheme of all the xrayutilities Gridders is pretty similar.
        if images.chunks is None:
            qcoord   = self._calc_qcoord(_ang2qObj, angles, _ang2qDict, gridSize) 
            grd      = self._calc_qmap(images, qcoord, gridSize, _gridderDict)
            grd_data = self._retr_qdata(grd)
        else:
            qcoord   = dask.delayed(self._calc_qcoord)(_ang2qObj, angles, _ang2qDict, gridSize) 
            grd      = dask.delayed(self._calc_qmap)(images, qcoord, gridSize, _gridderDict)
            grd_data = dask.array.from_delayed(dask.delayed(self._retr_qdata)(grd),
                                               shape=tuple([d[1] for d in gridSize.items()]),
                                               dtype=float)
        
        # ...the tricky part is creating the DataArrays. Specifically,
        # retrieving the q coordinates from the gridder. They are in `grd`
        # attributes called 'xaxis', 'yaxis', ... according to dimension.
        # We always use qx/qy/qz for dimension keys.
        coords = {}
        for i,axname in enumerate(gridSize.keys()):
            if images.chunks is None:
                axvals = self._retr_qaxis(grd, i)
            else:
                axvals = dask.array.from_delayed(
                    dask.delayed(self._retr_qaxis)(grd, i),
                    shape=(gridSize[d],),
                    dtype=float)
                
            # The xrayutilities Gridder will return a scalar instead of
            # a proper array of any of the dimensions that have length 1. But
            # we need a proper array to store this as a coordinate axis.
            coords[axname] = axvals if len(axvals.shape) > 0 else axvals[None,...]
    
        data_dims = tuple([k for k in coords.keys()])

        da = Dataset(data_vars={result_name: (data_dims, grd_data)},
                     coords=coords)

        return da


class LazyQMap(QMapper):
    ''' QMapper subclass which accepts data at initialization time.

    The idea is to follow up with a later call to a class
    instance (i.e. use the `.__call__()` operator) to trigger a
    Q-space mapping after some processing has been applied.
    To do this, `LazyQMap` stores the data in an internal `xarray.Dataset`
    of its own, the `.xdata` property.
    The raw / authoritative data (i.e. XRD images) is additionally
    accessed by the `.data` property, which behaves like a Python
    `dict()`, while `.angles` gives access to the desginated angles
    (goniometer and detector axes).
    
    The `.__init__()` method requires at least the experimental setup
    to be passed -- see documentation of `.__init__()` for details.
    
    All of the computed data being *lazily* evaluated means that any
    processing that must take place on the raw (i.e. untransformed)
    data can -- and must -- take place before first access to any
    of the `q...` properties. E.g. for intensity normalization, you
    could do simething like: `qmapper.xdata['img'] *= intensity` and
    only then proceed to accessing `.__call__()`.
    '''

    def __init__(self, setup=None, **data):
        ''' Initializes the Q-mapper with default settings, data, or both.

        Args:

            setup: This is expected to be an experiment definition dictionary,
              largely the same as the parameters of `QMapper.init_experiment()`.
              Additionally, this dicttionary also accepts the following keys:
                - `detectorTARAngles`: data for the tilt, azimuth and rotation angles.
                  Only required (and accepted) for angles that are defined as not `None`
                  in the `detectorTARAxes`. The parameter can be one of:
                   - A dictionary with angle name(s) as keys, and data array(s)
                     as values, for each of the directions: tilt, azimuth, and rotation.
                     Only directions with are marked with something different than `None`
                     in `detectorTARAxes` are accepted.   
                   - A tuple of strings (keys) for the corresponding angle names, if
                     the angle data is not supplied separately but is instead included
                     in the `data` container.
                - `goniometerAngles`: similarly to `detectorTARAngles`, this describes the
                  angles by which the goniometer can be positioned.
                  Can be one of:
                    - A dictionary with angle names as keys, and dara arrays as values,
                      from outer-most to inner-mot angle. *All* goniometer angles
                      named in `goniometerAxes` must be listed here.
                    - A tuple with strings, representing data-vars, if angles are
                      not supplied separatetly but within the `data` parameter itself.
                  Order is essential, it must be the same as the axis order in
                  `goniometerAxes`.

            **data: This is a series of named parameters, each containing data
              of the same length in the first dimension. This data can be either
              detector images, angles, or any kind of additional data.
        
        '''

        # Make a clean setup dicitonary which we can pass to superclass:
        #  - filter out angles data (goniometerAngles and detectorTARAngles
        #  - translate Axes (goniometerAxes and detectorTARAxes) to dict(),
        #    containing proper axes names
        #  - (take angle information and store it in the .xdata Dataset)
        
        s = { k:setup[k] for k in filter(lambda x: not x.endswith('Angles'), setup) }
        clean_setup = s.copy()
        
        for scope in ('goniometer', 'detectorTAR'):
            clean_setup[f'{scope}Axes'] = {
                k:v for k,v in zip(setup[f'{scope}Angles'].keys(),
                                   s[f'{scope}Axes'])
            }

        super().__init__(**clean_setup)

        # compile angle key list
        self._angle_keys = []
        for scope in ('goniometer', 'detectorTAR'):
            self._angle_keys += [k for k in setup[f'{scope}Angles'].keys()]

        

        self.xdata = self.__make_dataset(**data,
                                         **(setup['goniometerAngles']),
                                         **(setup['detectorTARAngles']))

        self._data_keys = tuple([k for k in data])


    @property
    def angles(self):
        ''' dict-based access to all the "angles" fields (mimics old API).
        '''
        return { k:self.xdata[k].values for k in self._angle_keys }


    @property
    def data(self):
        ''' dict-based access to all the "data" fields (mimics old API).
        '''
        return { k:self.xdata[k].values for k in self._data_keys }
    

    def __make_dataset(self, **dsets):
        ''' Creates an `xarray.Dataset` of data sets within `dsets`.
        The first dimension of all `dsets` is required to be the same.
        This is introduced as the first dimension in the `xarray` dataset,
        with the name "index".
        '''
        
        __arrayify = lambda x, n: x if hasattr(x, "__len__") else np.array([x]*n) 

        tmp = next(iter(dsets.items()))
        try:
            if not 'index' in dsets:
                dsets['index'] = np.array(range(tmp[1].shape[0]))

            xdata = Dataset(coords={'index': dsets['index']})
            for k,p in dsets.items():
                if k == 'index':
                    continue
                
                data = __arrayify(p, tmp[1].shape[0])
                dims = ["index"] + [f"{k}_{i}" for i in range(1,len(data.shape))]
                xdata[k] = (dims, data)

        except:
            logging.error(f"Error with dataset: {tmp}")
            raise

        return xdata

    
    def __getitem__(self, label):
        return self.xdata[label]


    def __call__(self, data_key,
                 qsize=None,
                 dims=None,
                 _gridderDict=None,
                 _ang2qDict=None):
        ''' Executes a Q-space mapping on the `data_key` array of the internal `.xdata`.

        Args:
        
            data_key: string with the data variable name to execute the mapping on

            qsize: Grid size(s) for the resulting Q-space map. Unlike the base class's
              `.qmap()` call, this supports only tuples of integers -- no dicts.

            dims: Dimension names to control the mapping. This is a tuple of combintations
              of "qx", "qy" and "qz".

            _gridderDict: update the `.gridderDict` before the mapping (i.e. a dictionary
              with extra parameters to pass to the gridder). This is strongly dependent
              on the unterlying (xrayutilities) implementation.

            _ang2qDict: update the `.ang2qDict` before the mapping (i.e. a dictionary
              with extra parameters to pass to the angular converter).
              This is strongly dependent on the unterlying (xrayutilities) implementation.

        Returns: an `xarray.Dataset` with detector data converted into Q-space,
          see also `super().qmap()` for details (...for which this method is
          typically just a wrapper).
        '''
        
        valid_kw = { 'qimg', 'dims', '_gridderDict', '_ang2qDict' }
        
        if _gridderDict is not None:
            self.setupGridder(**_gridderDict)

        if _ang2qDict is not None:
            self.setupAng2Q(**_ang2qDict)

        # LazyQMap also accepts simple x/y/z values, but QMapper does not.
        # Need to prepend a 'q' in front of single letters.
        if dims is not None:
            dims = tuple([ (f'q{x}' if len(x)==1 else x) for x in dims ])
            
        return self.qmap(self.xdata,
                         angles=self._angle_keys,
                         images=data_key,
                         qsize=qsize,
                         dims=dims,
                         retain="data_vars")


    
## Example for a class that does more than LazyQMap (namely accept angle offsets
## in its constructor), but still uses LazyQMap under the hood.
class OffsetQMap(LazyQMap):
    def __init__(self, offsets=None, **kwargs):
        super().__init__(**kwargs)
        if offsets is not None:
            for k,v in offsets.items():
                self.angles[k] += v
